@@ -22,7 +22,7 @@
 #define FUSE_MAJOR_VERSION 2
 
 /** Minor version of FUSE library interface */
-#define FUSE_MINOR_VERSION 7
+#define FUSE_MINOR_VERSION 8
 
 #define FUSE_MAKE_VERSION(maj, min)  ((maj) * 10 + (min))
 #define FUSE_VERSION FUSE_MAKE_VERSION(FUSE_MAJOR_VERSION, FUSE_MINOR_VERSION)
@@ -65,8 +65,12 @@ struct fuse_file_info {
 	    operation.	Introduced in version 2.6 */
 	unsigned int flush : 1;
 
+	/** Can be filled in by open, to indicate that the file is not
+	    seekable.  Introduced in version 2.9 */
+	unsigned int nonseekable : 1;
+
 	/** Padding.  Do not use*/
-	unsigned int padding : 29;
+	unsigned int padding : 28;
 
 	/** File handle.  May be filled in by filesystem in open().
 	    Available in all other file operations */
@@ -76,6 +80,38 @@ struct fuse_file_info {
 	uint64_t lock_owner;
 };
 
+/**
+ * Capability bits for 'fuse_conn_info.capable' and 'fuse_conn_info.want'
+ *
+ * FUSE_CAP_ASYNC_READ: filesystem supports asynchronous read requests
+ * FUSE_CAP_POSIX_LOCKS: filesystem supports "remote" locking
+ * FUSE_CAP_ATOMIC_O_TRUNC: filesystem handles the O_TRUNC open flag
+ * FUSE_CAP_EXPORT_SUPPORT: filesystem handles lookups of "." and ".."
+ * FUSE_CAP_BIG_WRITES: filesystem can handle write size larger than 4kB
+ * FUSE_CAP_DONT_MASK: don't apply umask to file mode on create operations
+ */
+#define FUSE_CAP_ASYNC_READ	(1 << 0)
+#define FUSE_CAP_POSIX_LOCKS	(1 << 1)
+#define FUSE_CAP_ATOMIC_O_TRUNC	(1 << 3)
+#define FUSE_CAP_EXPORT_SUPPORT	(1 << 4)
+#define FUSE_CAP_BIG_WRITES	(1 << 5)
+#define FUSE_CAP_DONT_MASK	(1 << 6)
+
+/**
+ * Ioctl flags
+ *
+ * FUSE_IOCTL_COMPAT: 32bit compat ioctl on 64bit machine
+ * FUSE_IOCTL_UNRESTRICTED: not restricted to well-formed ioctls, retry allowed
+ * FUSE_IOCTL_RETRY: retry with new iovecs
+ *
+ * FUSE_IOCTL_MAX_IOV: maximum of in_iovecs + out_iovecs
+ */
+#define FUSE_IOCTL_COMPAT	(1 << 0)
+#define FUSE_IOCTL_UNRESTRICTED	(1 << 1)
+#define FUSE_IOCTL_RETRY	(1 << 2)
+
+#define FUSE_IOCTL_MAX_IOV	256
+
 /**
  * Connection information, passed to the ->init() method
  *
@@ -109,14 +145,25 @@ struct fuse_conn_info {
 	 */
 	unsigned max_readahead;
 
+	/**
+	 * Capability flags, that the kernel supports
+	 */
+	unsigned capable;
+
+	/**
+	 * Capability flags, that the filesystem wants to enable
+	 */
+	unsigned want;
+
 	/**
 	 * For future use.
 	 */
-	unsigned reserved[27];
+	unsigned reserved[25];
 };
 
 struct fuse_session;
 struct fuse_chan;
+struct fuse_pollhandle;
 
 /**
  * Create a FUSE mountpoint
@@ -177,6 +224,13 @@ int fuse_daemonize(int foreground);
  */
 int fuse_version(void);
 
+/**
+ * Destroy poll handle
+ *
+ * @param ph the poll handle
+ */
+void fuse_pollhandle_destroy(struct fuse_pollhandle *ph);
+
 /* ----------------------------------------------------------- *
  * Signal handling					       *
  * ----------------------------------------------------------- */
