@@ -6,7 +6,7 @@
   See the file COPYING.LIB
 */
 
-#include "fuse_lowlevel.h"
+#include "fuse_i.h"
 #include "fuse_misc.h"
 #include "fuse_common_compat.h"
 #include "fuse_lowlevel_compat.h"
@@ -17,16 +17,6 @@
 #include <assert.h>
 #include <errno.h>
 
-struct fuse_session {
-	struct fuse_session_ops op;
-
-	void *data;
-
-	volatile int exited;
-
-	struct fuse_chan *ch;
-};
-
 struct fuse_chan {
 	struct fuse_chan_ops op;
 
@@ -121,6 +111,11 @@ int fuse_session_exited(struct fuse_session *se)
 		return se->exited;
 }
 
+void *fuse_session_data(struct fuse_session *se)
+{
+	return se->data;
+}
+
 static struct fuse_chan *fuse_chan_new_common(struct fuse_chan_ops *op, int fd,
 					      size_t bufsize, void *data,
 					      int compat)
