@@ -229,15 +229,20 @@ static void read_conf(void)
 					strip_line(line);
 					parse_line(line, linenum);
 				} else {
-					fprintf(stderr, "%s: reading %s: line %i too long\n",
-						progname, FUSE_CONF, linenum);
 					isnewline = 0;
 				}
-			} else if(line[strlen(line)-1] == '\n')
+			} else if(line[strlen(line)-1] == '\n') {
+				fprintf(stderr, "%s: reading %s: line %i too long\n", progname, FUSE_CONF, linenum);
+
 				isnewline = 1;
+			}
 			if (isnewline)
 				linenum ++;
 		}
+		if (!isnewline) {
+			fprintf(stderr, "%s: reading %s: missing newline at end of file\n", progname, FUSE_CONF);
+
+		}
 		fclose(fp);
 	} else if (errno != ENOENT) {
 		fprintf(stderr, "%s: failed to open %s: %s\n",
@@ -526,6 +531,7 @@ static int do_mount(const char *mnt, char **typep, mode_t rootmode,
 		*typep = type;
 		*mnt_optsp = mnt_opts;
 	}
+	free(fsname);
 	free(optbuf);
 
 	return res;
@@ -909,6 +915,10 @@ int main(int argc, char *argv[])
 	if (optind >= argc) {
 		fprintf(stderr, "%s: missing mountpoint argument\n", progname);
 		exit(1);
+	} else if (argc > optind + 1) {
+		fprintf(stderr, "%s: extra arguments after the mountpoint\n",
+			progname);
+		exit(1);
 	}
 
 	origmnt = argv[optind];
